#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <linux/posix_types.h>

#define DEVNAME			"/dev/dsp"
#define MIXER_DEVNAME		"/dev/mixer"
#define PARTITION_SEARCHSTRING	"root="

#define __SIOW          _IOW
#define __SIOWR         _IOWR
#define __SIOR          _IOR
#define __SIO           _IO

#define	  APF_NORMAL			0			/* Normal applications */
#define SNDCTL_DSP_PROFILE              __SIOW ('P', 23, int)   /* OBSOLETE */
#define SNDCTL_DSP_SETFMT               __SIOWR('P',5, int)     /* Selects ONE fmt */
#define SNDCTL_DSP_CHANNELS             __SIOWR('P', 6, int)
#define SNDCTL_DSP_SPEED                __SIOWR('P', 2, int)
#define SNDCTL_DSP_SYNC                 __SIO  ('P', 1)

#define WAVE_FORMAT_PCM                 0x0001
#define WAVE_FORMAT_ADPCM               0x0002
#define WAVE_FORMAT_ALAW                0x0006
#define WAVE_FORMAT_MULAW               0x0007
#define WAVE_FORMAT_IMA_ADPCM           0x0011

#define AFMT_QUERY       0x00000000      /* Return current fmt */
#define AFMT_MU_LAW      0x00000001
#define AFMT_A_LAW       0x00000002
#define AFMT_IMA_ADPCM   0x00000004
#define AFMT_U8          0x00000008
#define AFMT_S16_LE      0x00000010      /* Little endian signed 16 */
#define AFMT_S16_BE      0x00000020      /* Big endian signed 16 */
#define AFMT_S8          0x00000040
#define AFMT_U16_LE      0x00000080      /* Little endian U16 */
#define AFMT_U16_BE      0x00000100      /* Big endian U16 */
#define AFMT_MPEG        0x00000200      /* MPEG (2) audio */
#define AFMT_S32_LE      0x00001000


#define MIXER_WRITE(dev)        __SIOWR('M', dev, int) /* Mixer with a big M */
#define MIXER_READ(dev)         __SIOR('M', dev, int)  /* and his reading friend */
#define SOUND_MIXER_DEVMASK     0xfe

typedef struct
{
  int coeff1, coeff2;
}
adpcm_coeff;


static int
le_int (unsigned char *p, int l)
{
  int i, val;

  val = 0;

  for (i = l - 1; i >= 0; i--)
  {
    val = (val << 8) | p[i];
  }

  return val;
}

#if 0
static int
be_int (unsigned char *p, int l)
{
  int i, val;

  val = 0;

  for (i = 0; i < l; i++)
  {
    val = (val << 8) | p[i];
  }

  return val;
}
#endif /* 0 */
static int dump_data (int fd, int audiofd, int filesize)
{
  int bsize, l;
  unsigned char buf[1024];

  bsize = sizeof (buf);

  while (filesize)
  {
    l = bsize;

    if (l > filesize)
      l = filesize;

    if ((l = read (fd, buf, l)) <= 0)
    {
      return l;
    }

    if (write (audiofd, buf, l) == -1)
    {
      perror (DEVNAME);
      return -1;
    }

    filesize -= l;
  }
  return 0;
}

static int
dump_data_24 (int fd, int audiofd, int filesize)
{
  /*
  * Dump 24 bit packed audio data to the device (after converting to 32 bit).
  */
  int bsize, i, l;
  unsigned char buf[1024];
  int outbuf[1024], outlen = 0;

  int sample_s32;

  bsize = sizeof (buf);

  filesize -= filesize % 3;

  while (filesize >= 3)
  {
    l = bsize - bsize % 3;
    if (l > filesize)
      l = filesize;

    if (l < 3)
      break;

    if ((l = read (fd, buf, l)) <= 0)
    {
      return l;
    }

    outlen = 0;

    for (i = 0; i < l; i += 3)
    {
      unsigned int *u32 = (unsigned int *) &sample_s32;	/* Alias */

      /* Read the litle endian input samples */
      *u32 = (buf[i] << 8) | (buf[i + 1] << 16) | (buf[i + 2] << 24);
      outbuf[outlen++] = sample_s32;

    }

    if (write (audiofd, outbuf, outlen * sizeof (int)) == -1)
    {
      perror (DEVNAME);
      return -1;
    }

    filesize -= l;
  }
  return 0;
}

int set_volume(int oss_mixer_number,int volume)
{
	/* oss_mixer_number => include/linux/soundcard.h -> SOUND_MIXER_VOLUME */
	int fmixer =-1;
	int mask =-1;
	int oss_volume =-1;

	/* open the mixer */
	if ( (fmixer = open(MIXER_DEVNAME, O_RDWR)) == -1){
		perror(MIXER_DEVNAME);
		return -1;
	}

	/* read the oss device mask to see if the oss_mixer_number 
	 * is in the list of usable mixers */
        if (ioctl(fmixer, MIXER_READ(SOUND_MIXER_DEVMASK), &mask)){
		fprintf(stderr,"Failed to read the mixer dev mask\n");
                perror(MIXER_DEVNAME);
		close(fmixer);
		return -1;
        }

	if( mask & (1 << oss_mixer_number)) {
		/* oss volume is stored in 2 bytes that will contain
		 * a value between 0x00 and 0x64 (100 decimal)
		 */
		oss_volume =  ( volume & 0x7F ) |   (volume & 0x7F ) << 8;
		if (ioctl(fmixer, MIXER_WRITE(oss_mixer_number), &oss_volume)){
			perror("Set volume");
			close(fmixer);
			return -1;
		}
	} else {
		fprintf(stderr,"The oss mixer with number %i is not configured mask( 0x%08X )\n",oss_mixer_number,mask);
		close(fmixer);
		return -1;
	}
	close(fmixer);
	return 0;
}

static int
setup_device (int fd, int audiofd, int channels, int bits, int speed)
{
  int tmp;

  tmp = APF_NORMAL;
  ioctl (audiofd, SNDCTL_DSP_PROFILE, &tmp);

  tmp = bits;

  if (ioctl (audiofd, SNDCTL_DSP_SETFMT, &tmp) == -1)
  {
    perror (DEVNAME);
    fprintf (stderr, "Failed to select bits/sample\n");
    return 0;
  }

  if (tmp != bits)
  {
    fprintf (stderr, "%s doesn't support this audio format (%x/%x).\n",
             DEVNAME, bits, tmp);
    return 0;
  }

  tmp = channels;

  if (ioctl (audiofd, SNDCTL_DSP_CHANNELS, &tmp) == -1)
  {
    perror (DEVNAME);
    fprintf (stderr, "Failed to select number of channels.\n");
    return 0;
  }

  if (tmp != channels)
  {
    fprintf (stderr, "%s doesn't support %d channels.\n",
             DEVNAME, channels);
    return 0;
  }

  tmp = speed;

  if (ioctl (audiofd, SNDCTL_DSP_SPEED, &tmp) == -1)
  {
    perror (DEVNAME);
    fprintf (stderr, "Failed to select sampling rate.\n");
    return 0;
  }

  return 1;
}


static int
play_ms_adpcm_wave (char *filename, int fd, int audiofd , unsigned char *hdr, int l)
{
  int i, n, dataleft, x;

  adpcm_coeff coeff[32];
  static int AdaptionTable[] = 	{ 230, 230, 230, 230, 307, 409, 512, 614,
                                   768, 614, 512, 409, 307, 230, 230, 230
  				};

  unsigned char buf[4096];

  int channels = 1;
  int bits = 8;
  int speed = 11025;
  int p = 12;
  int max;
  int outp = -1;
  int nBlockAlign = -1;
  int wSamplesPerBlock = -1;
  int wNumCoeff = -1;
  int fmt = -1;
  int nib;

  unsigned char outbuf[64 * 1024];

  /* filelen = le_int (&hdr[4], 4); */

  while (p < l - 16 && memcmp (&hdr[p], "data", 4) != 0)
  {
    n = le_int (&hdr[p + 4], 4);

    if (memcmp (&hdr[p], "fmt ", 4) == 0)
    {
      fmt = le_int (&hdr[p + 8], 2);
      channels = le_int (&hdr[p + 10], 2);
      speed = le_int (&hdr[p + 12], 4);
      nBlockAlign = le_int (&hdr[p + 20], 2);
      /* bytes_per_sample = le_int (&hdr[p + 20], 2); */

      bits = AFMT_S16_LE;

      wSamplesPerBlock = le_int (&hdr[p + 26], 2);
      wNumCoeff = le_int (&hdr[p + 28], 2);

      x = p + 30;

      for (i = 0; i < wNumCoeff; i++)
	    {
	      coeff[i].coeff1 = (short) le_int (&hdr[x], 2);
	      x += 2;
	      coeff[i].coeff2 = (short) le_int (&hdr[x], 2);
	      x += 2;
	    }

    }

    p += n + 8;
  }

  if (p < l - 16 && memcmp (&hdr[p], "data", 4) == 0)
  {

    dataleft = n = le_int (&hdr[p + 4], 4);
    p += 8;

    if (lseek (fd, p, SEEK_SET) == -1)
    {
      perror (filename);
      fprintf (stderr, "Can't seek to the data chunk\n");
      return -1;
    }

    if (fmt != WAVE_FORMAT_ADPCM)
    {
      fprintf (stderr, "%s: Unsupported wave format %04x\n", filename,
               fmt);
      return -2;
    }

    if (!setup_device (fd, audiofd, channels, bits, speed))
      return -3;

    /*
    * Playback procedure
    */
#define OUT_SAMPLE(s) { \
                          if (s>32767)s=32767;else if(s<-32768)s=-32768; \
                            outbuf[outp++] = (unsigned char)(s & 0xff); \
                              outbuf[outp++] = (unsigned char)((s>>8) & 0xff); \
                                n+=2; \
                                  if (outp>=max){write(audiofd, outbuf, outp);outp=0;}\
}

#define GETNIBBLE \
    ((nib==0) ? \
     (buf[x + nib++] >> 4) & 0x0f : \
     buf[x++ + --nib] & 0x0f \
     )

#if 0
      /*
      * There is no idea in using SNDCTL_DSP_GETBLKSIZE in applications like this.
      * Using some fixed local buffer size will work equally well.
      */
      if (ioctl (audiofd, SNDCTL_DSP_GETBLKSIZE, &max) == -1)
        perror ("SNDCTL_DSP_GETBLKSIZE");
#else
    max = 1024;
#endif
    outp = 0;

    while (dataleft > nBlockAlign &&
           read (fd, buf, nBlockAlign) == nBlockAlign)
    {
      int predictor[2], delta[2], samp1[2], samp2[2];

      int x = 0;
      dataleft -= nBlockAlign;

      nib = 0;
      n = 0;

      for (i = 0; i < channels; i++)
	    {
	      predictor[i] = buf[x];
	      x++;
	    }

      for (i = 0; i < channels; i++)
	    {
	      delta[i] = (short) le_int (&buf[x], 2);
	      x += 2;
	    }

      for (i = 0; i < channels; i++)
	    {
	      samp1[i] = (short) le_int (&buf[x], 2);
	      x += 2;
	      OUT_SAMPLE (samp1[i]);
	    }

      for (i = 0; i < channels; i++)
	    {
	      samp2[i] = (short) le_int (&buf[x], 2);
	      x += 2;
	      OUT_SAMPLE (samp2[i]);
	    }

      while (n < (wSamplesPerBlock * 2 * channels))
        for (i = 0; i < channels; i++)
	      {
          int pred, new, error_delta, i_delta;

          pred = ((samp1[i] * coeff[predictor[i]].coeff1)
                  + (samp2[i] * coeff[predictor[i]].coeff2)) / 256;
          i_delta = error_delta = GETNIBBLE;

          if (i_delta & 0x08)
            i_delta -= 0x10;	/* Convert to signed */

          new = pred + (delta[i] * i_delta);
          OUT_SAMPLE (new);

          delta[i] = delta[i] * AdaptionTable[error_delta] / 256;
          if (delta[i] < 16)
            delta[i] = 16;

          samp2[i] = samp1[i];
          samp1[i] = new;
	      }
    }

  }

  if (outp > 0)
    write (audiofd, outbuf, outp /*(outp+3) & ~3 */ );
  return 0;
}

int play_wave (char *filename, int fd, int audiofd, unsigned char *hdr, int l)
{
  int filelen;

  int n;

  int channels = 1;
  int bits = 8;
  int file_bits = 8;
  int speed = 11025;
  int p = 12;
  int fmt = -1;
  int retval;
  filelen = le_int (&hdr[4], 4);

  while (p < l - 16 && memcmp (&hdr[p], "data", 4) != 0)
  {
    n = le_int (&hdr[p + 4], 4);

    if (memcmp (&hdr[p], "fmt ", 4) == 0)
    {
      fmt = le_int (&hdr[p + 8], 2);
      channels = le_int (&hdr[p + 10], 2);
      speed = le_int (&hdr[p + 12], 4);
      /* bytes_per_sec = le_int (&hdr[p + 16], 4); */
      /* bytes_per_sample = le_int (&hdr[p + 20], 2); */

      bits = le_int (&hdr[p + 22], 2);

      if (fmt == WAVE_FORMAT_ADPCM)
	    {
	      return play_ms_adpcm_wave (filename, fd, audiofd, hdr, l);
	    }
    }

    p += n + 8;
  }

  if (p < l - 16 && memcmp (&hdr[p], "data", 4) == 0)
  {

    n = le_int (&hdr[p + 4], 4);
    p += 8;

    if (lseek (fd, p, SEEK_SET) == -1)
    {
      perror (filename);
      fprintf (stderr, "Can't seek to the data chunk\n");
      return -1;
    }

    if (fmt != WAVE_FORMAT_PCM)
      switch (fmt)
      {
      case WAVE_FORMAT_ALAW:
        bits = AFMT_A_LAW;
        break;

      case WAVE_FORMAT_MULAW:
        bits = AFMT_MU_LAW;
        break;

      case WAVE_FORMAT_IMA_ADPCM:
        bits = AFMT_IMA_ADPCM;
        break;

      default:
        fprintf (stderr, "%s: Unsupported wave format %04x\n", filename,
                 fmt);
        return -2;
      }

    file_bits = bits;

    if (bits == 32)
    {
      bits = AFMT_S32_LE;
    }
    else if (bits == 24)
    {
      bits = AFMT_S32_LE;
    }
    if (!setup_device (fd, audiofd, channels, bits, speed))
      return -3;

    if (file_bits != 24)
    {
      retval=dump_data (fd, audiofd, n);
    }
    else
    {
      retval=dump_data_24 (fd, audiofd, n);
    }
    if( retval < 0 )
	return retval;
  }
  return 0;
}

void snd_dsp_sync(int audio_handle,int value){
  ioctl (audio_handle, SNDCTL_DSP_SYNC, value);
};

int play (char * pWave)
{
  int fd;
  int l;
  int audiofd;
  int exitstatus = 0;
  unsigned char buf[1024];

  if ((audiofd = open (DEVNAME, O_WRONLY, 0)) == -1)
  {
    perror (DEVNAME);
    exitstatus++;
    return exitstatus;
  }

  if ((fd = open (pWave, O_RDONLY, 0)) == -1)
  {
    perror (pWave);
    exitstatus++;
    return exitstatus;
  }


  if ((l = read (fd, buf, sizeof (buf))) == -1)
  {
    perror (pWave);
    exitstatus++;
    close (fd);
    return exitstatus;
  }

  if (l == 0)
  {
    fprintf (stderr, "%s is empty file.\n", pWave);
    exitstatus++;
    close (fd);
    return exitstatus;
  }

  lseek (fd, 0, SEEK_SET);	/* Start from the beginning */

  if (l > 16 && memcmp (&buf[0], "RIFF", 4) == 0 && memcmp (&buf[8], "WAVE", 4) == 0)
  {
    if( play_wave (pWave, fd, audiofd, buf, l) < 0 )
    {
        exitstatus++;
	return exitstatus;
    }
  }
  else
  {
    exitstatus++;
    close (fd);
    return exitstatus;
  }

  close (fd);

  snd_dsp_sync(audiofd,0);

  close (audiofd);
  return exitstatus;
}

